import java.io.*;
import java.util.*;

public class Main
{
	public static void main( String[] args )
	{
		if( args.length != 1 )
		{
			System.err.println("Usage:");
			System.err.println("  java -jar mtranc.jar [template.mtt]");
		}
		else try
		{
			compile( args[0] );
			System.err.println();
			System.err.println("  ** OK **");
		}
		catch( Exception e )
		{
			System.err.println("error:");
			System.err.println("  ## "+e+" ##");
		}
	}

	static void compile( String src ) throws Exception
	{
		File fp = new File(src);
		Parser.Program p = (new Parser()).parseProgram(fp, null);

		String dest = src.substring(0, src.lastIndexOf('.'));
		File dir = new File(dest);
		if( !dir.isDirectory() && !dir.mkdir() )
			throw new Exception("Directory \""+dest+"\" cannot be created.");

		translate( p, dir );
	}

	static void translate( Parser.Program p, File dir ) throws Exception
	{
		File mona = new File(dir, "main.mona"); mona.createNewFile();
		File tmpl = new File(dir, "template");  tmpl.createNewFile();

		OutputStreamWriter mp =
			new OutputStreamWriter(new FileOutputStream(mona));
		OutputStreamWriter tp =
			new OutputStreamWriter(new FileOutputStream(tmpl));

		try {
			List<String[]> eqs = eq_process(p);
			process(p, mp, tp, eqs);
		} finally {
			mp.close();
			tp.close();
		}

		run_mona( mona.getName(), dir );
	}

	static int eq_p_counter = 0;
	static String toImage( Parser.Node p )
	{
		if( p instanceof Parser.Token )
			return ((Parser.Token)p).getOriginalImage();

		String buf = "";
		for( Parser.Node n : p.getChildNodes() )
			buf += toImage(n);
		return buf;
	}

	static List<String[]> eq_process( Parser.Program p ) throws Exception
	{
		final List<String[]> ans = new ArrayList<String[]>();
/*
		(new Parser.Visitor() {
			public void visitAll( Parser.Node node ) {
				for( Object o : node.getChildNodes() ) {
					Parser.Node child = (Parser.Node) o;
					if( child instanceof Parser.MSO_Op ) {
						Parser.MSO_Op e = (Parser.MSO_Op) o;
						if( e.op().getImage().equals("eq") ) {
							String[] v = new String[]{"__q" + eq_p_counter   +"a",
							                          "__q" + eq_p_counter++ +"b"};
							ans.add(v);
							String text = toImage(e.t1()) + " = " + v[0]
							    + " & " + v[1] + " =" + toImage(e.t2());
							try {
							Parser.MSO_And newChild =
								(new Parser()).parseMSO_And("temp", text, 8);
							((Parser.Default.Node)node).replaceChild(child, newChild);
							} catch( Parser.ParseException _ ) {}
						}
					}
					visitAll(child);
				}
			}
		}).visitAll(p);
*/
		return ans;
	}

	static void process( Parser.Program p,
		OutputStreamWriter mona, OutputStreamWriter tmpl,
		List<String[]> eqs ) throws Exception
	{
		//---------------------------------------
		// Phase 0 : Primitive Predicates
		//---------------------------------------
		final StringBuilder pr = new StringBuilder();
		{
			pr.append( "m2l-tree;\n" );
			pr.append( "pred p_cs( var1 parent, var2 children ) =\n" );
			pr.append( "  all1 c: (c in children <=>\n" );
			pr.append( "      (c = parent.0) | (ex1 b: b.1=c & b in children));\n" );
			pr.append( "pred p_c( var1 parent, var1 child ) =\n" );
			pr.append( "  ex2 children: (p_cs(parent,children) & child in children);\n" );
			pr.append( "pred a_d( var1 ancestor, var1 descendant ) =\n" );
			pr.append( "  ancestor.0 <= descendant;\n" );
			pr.append( "pred p_n( var1 psib, var1 nsib ) =\n" );
			pr.append( "  psib<nsib & all1 parent: (p_c(parent,psib) => p_c(parent,nsib));\n" );
			pr.append( "pred d_order( var1 p, var1 q ) =\n" );
			pr.append( "   a_d(p,q) | ex1 r: (a_d(r,p)|r=p) & r.1<=q;\n" );
		}

		//---------------------------------------
		// Phase 1 : Collect Label Variables
		//---------------------------------------
		final StringBuilder labvar  = new StringBuilder();
		final StringBuilder realvar = new StringBuilder();
		{
			final Set<String> added = new HashSet<String>();

			(new Parser.Visitor() {
				protected void visit( Parser.Gather g ) {
					String s = g.v().getImage();
					if( !added.contains(s) )
					{
						realvar.append("var1 ");
						realvar.append( s );
						realvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Visit v ) {
					String s = v.v().getImage();
					if( !added.contains(s) )
					{
						realvar.append("var1 ");
						realvar.append( s );
						realvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Query q ) {
					String s = "_";
					if( !added.contains(s) )
					{
						labvar.append("var1 ");
						labvar.append( s );
						labvar.append(";\n");
						added.add(s);
					}
				}


				protected void visit( Parser.Term_Atr atr ) {
					String s = atr.name().getImage();
					if( !added.contains(s) )
					{
						labvar.append("var2 ");
						term_translate( labvar, atr );
						labvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Term_Awild atr ) {
					String s = atr.name().getImage();
					if( !added.contains(s) )
					{
						labvar.append("var2 ");
						term_translate( labvar, atr );
						labvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Term_Tag tag ) {
					String s = tag.name().getImage();
					if( !added.contains(s) )
					{
						labvar.append("var2 ");
						term_translate( labvar, tag );
						labvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Term_Twild tag ) {
					String s = tag.name().getImage();
					if( !added.contains(s) )
					{
						labvar.append("var2 ");
						term_translate( labvar, tag );
						labvar.append(";\n");
						added.add(s);
					}
				}
				protected void visit( Parser.Term_Swild str ) {
					String s = str.name().getImage();
					if( !added.contains(s) )
					{
						labvar.append("var2 ");
						term_translate( labvar, str );
						labvar.append(";\n");
						added.add(s);
					}
				}
			}).visitAll(p);
		}

		//---------------------------------------
		// Phase  : Dummy Equality Variables
		//---------------------------------------
		for( String[] ss : eqs )
			for( String s : ss ) {
				realvar.append("var1 ");
				realvar.append(s);
				realvar.append(";\n");
			}

		//---------------------------------------
		// Phase 2: Library Definition Extraction
		//---------------------------------------
		final StringBuilder preds = new StringBuilder();
		{
			(new Parser.Visitor() {
				protected void visit( Parser.PredDef pd ) {
					HashSet<String> so = new HashSet<String>();

					preds.append("pred ");
					preds.append(pd.name().getImage());
					preds.append("(");
					boolean first = true;
					for( Parser.Param po : pd.param() ) {
						if(first)first=false;
						else preds.append( ", " );
						preds.append( po.type().getImage() );
						preds.append( " " );
						preds.append( po.name().getImage() );
						if( po.type().getImage() == "var2" )
							so.add(po.name().getImage());
					}
					preds.append(") = \n  ");
					mso_translate( preds, pd.q(), so );
					preds.append(";\n");
				}
			}).visitAll(p);
		}

		//---------------------------------------
		// Phase 3: Automata Extraction
		// Phase 4. Template ByteCode Generation
		//---------------------------------------
		final StringBuilder atmt = new StringBuilder();
		final StringBuilder code = new StringBuilder();
//*
		{
			(new Parser.Visitor() {
				public void visitAll(Parser.Node n) { visitNode(n); }

				protected void visit( Parser.Program e ) {
					visitNode( e.expr() );
				}

				protected void visit( Parser.ExprList e ) {
					for( Parser.Expr o : e.e() )
						visitNode( o );
				}

				protected void visit( Parser.Dyn e ) {
					visitNode( e.dyn() );
				}

				protected void visit( Parser.Elem e ) {
					code.append("e ");
					code.append( e.name().getImage() );
					code.append(" {");
					for( Parser.Expr o : e.expr().e() )
						visitNode( o );
					code.append(" }\n");
				}

				protected void visit( Parser.Attr e ) {
					code.append("a ");
					code.append( e.name().getImage().substring(1) );
					code.append(" {");
					for( Parser.Expr o : e.expr().e() )
						visitNode( o );
					code.append(" }\n");
				}

				protected void visit( Parser.DynElem e ) {
					code.append("E ");
					code.append( e.var().getImage() );
					code.append(" {");
					for( Parser.Expr o : e.expr().e() )
						visitNode( o );
					code.append(" }\n");
				}

				protected void visit( Parser.DynAttr e ) {
					code.append("A ");
					code.append( e.var().getImage() );
					code.append(" {");
					for( Parser.Expr o : e.expr().e() )
						visitNode( o );
					code.append(" }\n");
				}

				protected void visit( Parser.Pcdata e ) {
					code.append("s ");
					code.append( e.str().getImage() );
					code.append("\n");
				}

				protected void visit( Parser.Gather e ) {
					code.append("g ");
					code.append( e.v().getImage() );
					code.append(" {");
					Iterator<? extends Parser.MSO>      iq = e.q().iterator();
					Iterator<? extends Parser.ExprList> ie = e.e().iterator();
					while( iq.hasNext() ) {
						Parser.MSO      q = iq.next();
						Parser.ExprList el= ie.next();
						int id = automata_expr( atmt, q );
						code.append( id );
						code.append( " " );
						visitNode(el);
					}
					code.append(" }\n");
				}

				protected void visit( Parser.Visit e ) {
					code.append("v ");
					code.append( e.v().getImage() );
					code.append(" ");
					if( e.vf() != null )
						code.append( e.vf().getImage() );
					else
						code.append("-");
					code.append(" {");
					Iterator<? extends Parser.MSO>      iq = e.q().iterator();
					Iterator<? extends Parser.ExprList> ie = e.e().iterator();
					while( iq.hasNext() ) {
						Parser.MSO      q = iq.next();
						Parser.ExprList el= ie.next();
						int id = automata_expr( atmt, q );
						code.append( id );
						code.append( " " );
						visitNode(el);
					}
					code.append(" }\n");
				}

				protected void visit( Parser.Var e ) {
					code.append("X ");
					code.append( e.v().getImage() );
					code.append("\n");
				}

				protected void visit( Parser.Query e ) {
					code.append("g _ {");
					{
						Parser.MSO q = e.q();
						int id = automata_expr( atmt, q );
						code.append( id );
						code.append( " X _" );
					}
					code.append(" }\n");
				}
			}).visitAll(p);
		}
		mona.write(pr.toString());
		mona.write("\n\n\n");
		mona.write(realvar.toString());
		mona.write(labvar.toString());
		mona.write("\n\n\n");
		mona.write(preds.toString());
		mona.write("\n\n\n");
		mona.write(atmt.toString());
		tmpl.write(code.toString());
	}

	static void mso_translate(
		final StringBuilder sb, Parser.MSO mso, final HashSet<String> so )
	{
		(new Parser.Visitor() {
			public void visitAll(Parser.Node n) { visitNode(n); }

			protected void visit( Parser.MSO_Ex1 e ) {
				for( Parser.Token o : e.v() ) {
					sb.append("(ex1 ");
					sb.append( o.getImage() );
					sb.append(": ");
				}
				mso_translate( sb, e.q(), so );
				for( Parser.Token o : e.v() )
					sb.append(")");
			}
			protected void visit( Parser.MSO_All1 e ) {
				for( Parser.Token o : e.v() ) {
					sb.append("(all1 ");
					sb.append( o.getImage() );
					sb.append(": ");
				}
				mso_translate( sb, e.q(), so );
				for( Parser.Token o : e.v() )
					sb.append(")");
			}
			protected void visit( Parser.MSO_Ex2 e ) {
				HashSet<String> so2 = new HashSet<String>(so);
				for( Parser.Token o : e.v() ) {
					sb.append("(ex2 ");
					sb.append( o.getImage() );
					so.add(    o.getImage() );
					sb.append(": ");
				}
				mso_translate( sb, e.q(), so2 );
				for( Parser.Token o : e.v() )
					sb.append(")");
			}
			protected void visit( Parser.MSO_All2 e ) {
				HashSet<String> so2 = new HashSet<String>(so);
				for( Parser.Token o : e.v() ) {
					sb.append("(all2 ");
					sb.append( o.getImage() );
					so.add( o.getImage() );
					sb.append(": ");
				}
				mso_translate( sb, e.q(), so2 );
				for( Parser.Token o : e.v() )
					sb.append(")");
			}
			void binop( Parser.MSO q1, String op, Parser.MSO q2 ) {
				sb.append( "(" );
				mso_translate( sb, q1, so );
				sb.append( " "+op+" " );
				mso_translate( sb, q2, so );
				sb.append( ")" );
			}
			protected void visit( Parser.MSO_Then e ) { binop(e.q1(),"=>",e.q2()); }
			protected void visit( Parser.MSO_Iff e ) { binop(e.q1(),"<=>",e.q2()); }
			protected void visit( Parser.MSO_And e ) { binop(e.q1(),"&",e.q2()); }
			protected void visit( Parser.MSO_Or e ) { binop(e.q1(),"|",e.q2()); }
			protected void visit( Parser.MSO_Not e ) {
				sb.append( "(~" );
				mso_translate( sb, e.q(), so );
				sb.append( ")" );
			}
			protected void visit( Parser.MSO_Op e ) {
				term_translate(sb, (Parser.Term)e.t1());
				String op = e.op().getImage();
				sb.append( " "+op+" " );
				term_translate(sb, (Parser.Term)e.t2());
			}
			protected void visit( Parser.MSO_Const e ) {
				sb.append( e.str().getImage() );
			}
			protected void visit( Parser.MSO_Appl e ) {
				sb.append( e.pred().getImage() );
				sb.append( "(" );
				boolean first = true;
				for( Parser.Term o : e.arg() ) {
					if(first)first=false;
					else sb.append(",");
					term_translate(sb, o);
				}
				sb.append( ")" );
			}
			protected void visit( Parser.MSO_Path e ) {
				sb.append("(");
				String prev_var = null;
				int the_op = e.root()!=null ? e.root().getImage().equals("/")?1:0 : 0;
				Iterator<? extends Parser.Node>  u = e.unit().iterator();
				Iterator<? extends Parser.Token> a = e.axis().iterator();
				boolean first = true;
				while( u.hasNext() ) {
					Parser.Node unit = u.next();
					// ݂unitɑΉ1stOrderϐ
					String cur_var = null;
					String cur_set = null;
					{
						if( unit instanceof Parser.Term_Var ) {
							cur_var = ((Parser.Term_Var)unit).name().getImage();
							if( so.contains(cur_var) ) {
								cur_set = cur_var;
								cur_var = freshVar();
								if(first)first=false;else sb.append(" & ");
								sb.append("ex1 "+cur_var+":"+cur_var
								         +" in "+cur_set);
							}
						} else if( unit instanceof Parser.UnitIn ) {
							cur_var = ((Parser.UnitIn)unit).e().name().getImage();
							cur_set = str_of( ((Parser.UnitIn)unit).s() );
							if(first)first=false;else sb.append(" & ");
							sb.append(cur_var+" in "+cur_set);
						} else if( unit instanceof Parser.UnitNotIn ) {
							cur_var = ((Parser.UnitNotIn)unit).e().name().getImage();
							cur_set = str_of( ((Parser.UnitNotIn)unit).s() );
							if(first)first=false;else sb.append(" & ");
							sb.append(cur_var+" notin "+cur_set);
						} else {
							cur_var = freshVar();
							cur_set = str_of( (Parser.Term)unit );
							if(first)first=false;else sb.append(" & ");
							sb.append("ex1 "+cur_var+":"+cur_var
							         +" in "+cur_set);
						}
					}
					// relation  assertion ǉ
					if( prev_var != null ) {
						if(first)first=false;else sb.append(" & ");
						sb.append( the_op==1 ? "p_c" : "a_d" );
						sb.append( "("+ prev_var+","+cur_var+")" );
					} else if( prev_var == null && the_op==1 ) {
						if(first)first=false;else sb.append(" & ");
						sb.append( "root = "+cur_var );
					}
					prev_var = cur_var;
					if( a.hasNext() )
						the_op = a.next().getImage().equals("/")?1:2;
				}
				sb.append( ")" );
			}
		}).visitAll(mso);
	}

	static int varCounter = 0;
	static String freshVar() { return "__f" + ++varCounter; }
	static String str_of( Parser.Term t ) {
		StringBuilder sb = new StringBuilder();
		term_translate( sb, t );
		return sb.toString();
	}
	static void term_translate(
		final StringBuilder sb, Parser.Term t )
	{
		(new Parser.Visitor() {
			public void visitAll(Parser.Node n) { visitNode(n); }

			protected void visit( Parser.Term_Var t )
				{ sb.append(t.name().getImage()); }
			protected void visit( Parser.Term_VarM t )
				{ sb.append(t.name().getImage()); sb.append(t.mod().getImage());}
			protected void visit( Parser.Term_Atr t )
				{ sb.append("__a"); String s = t.name().getImage();
				  sb.append(s.substring(1)); }
			protected void visit( Parser.Term_Tag t )
				{ sb.append("__e"); String s = t.name().getImage();
				  sb.append(s.substring(1,s.length()-1)); }
			protected void visit( Parser.Term_Awild t )
				{ sb.append("__a"); }
			protected void visit( Parser.Term_Twild t )
				{ sb.append("__e"); }
			protected void visit( Parser.Term_Swild s )
				{ sb.append("__t"); }
			protected void visit( Parser.Term_Op t )
				{ term_translate(sb,t.t1());
				  sb.append(t.op().getImage());
				  term_translate(sb,t.t2()); }
		}).visitAll(t);
	}

	static int am_id = 0;
	static int automata_expr( StringBuilder sb, Parser.MSO mso )
	{
		int id = ++am_id;
		sb.append( "execute export( \"" );
		sb.append( id );
		sb.append( "\", " );
		HashSet<String> so = new HashSet<String>();
		mso_translate( sb, mso, so );
		sb.append( " );\n" );
		return id;
	}

	static void run_mona( String monaFile, File dir ) throws Exception
	{
		ProcessBuilder pb =
			new ProcessBuilder( "mona", "-n", "-u", monaFile );
		pb.directory(dir);
		pb.redirectErrorStream(true);
		Process p = pb.start();

		InputStream is = p.getInputStream();
		InputStreamReader isr = new InputStreamReader(is);
		BufferedReader bfr = new BufferedReader(isr);
		boolean err = false;
		for(;;) {
			String str = bfr.readLine();
			if( str == null )
				break;
			if( str.equals("Execution aborted") )
				err = true;
			if( str.length()>3 && (str.charAt(1)=='%'||str.charAt(2)=='%') )
				System.err.print(str+"\r");
			else if( str.trim().length() != 0 )
				System.err.println(str);
		}

		int result = p.waitFor();
		if( err )
			throw new Exception("Automata Compilation Error");
	}
}
