#include "common.h"
//#define NEW_ALGORTIHM

//========================================================================
// Lazy Sets
//========================================================================

typedef unsigned int SET_T;

struct InXMLSet : mem_man<InXMLSet>
{
	virtual void as_vec( int N, vector< vector<InXML*> >& r ) = 0;
};

InXMLSet* Union( InXMLSet* lhs, InXMLSet* rhs )
{
	struct impl : InXMLSet {
		InXMLSet *lhs, *rhs;
		impl(InXMLSet* l, InXMLSet* r): lhs(l), rhs(r) {}
		void as_vec( int N, vector< vector<InXML*> >& r ) {
			lhs->as_vec(N, r);
			rhs->as_vec(N, r);
		}
	};
	return !lhs ? rhs : !rhs ? lhs : new impl(lhs, rhs);
}

InXMLSet* Prod( SET_T li, InXMLSet* lhs, SET_T ri, InXMLSet* rhs )
{
	struct impl : InXMLSet {
		InXMLSet *lhs, *rhs;
		impl(InXMLSet* l, InXMLSet* r): lhs(l), rhs(r) {}
		void as_vec( int N, vector< vector<InXML*> >& r ) {
			vector< vector<InXML*> > ls, rs;
			lhs->as_vec(N, ls);
			rhs->as_vec(N, rs);
			for(int i=0; i!=ls.size(); ++i)
			for(int j=0; j!=rs.size(); ++j) {
				vector<InXML*> x(N);
				for(int k=0; k!=N; ++k)
					x[k] = ls[i][k] ? ls[i][k] : rs[j][k];
				r.push_back(x);
			}
		}
	};
	return !li ? rhs : !ri ? lhs : !lhs ? 0 : !rhs ? 0 : new impl(lhs, rhs);
}

InXMLSet* Prod( SET_T li, InXMLSet* lhs, SET_T ci, InXMLSet* chs, SET_T ri, InXMLSet* rhs )
{
	return Prod( li, lhs, ci|ri, Prod(ci, chs, ri, rhs) );
}

InXMLSet* Single( SET_T t, InXML* ix )
{
	struct impl : InXMLSet {
		SET_T  t;
		InXML* ix;
		impl(SET_T t, InXML* ix) : t(t), ix(ix) {}
		void as_vec( int N, vector< vector<InXML*> >& r ) {
			vector<InXML*> x(N);
			for(int k=0; k!=N; ++k)
				if( t & (1<<k) )
					x[k] = ix;
			r.push_back(x);
		}
	};
	return new impl(t, ix);
}

//========================================================================
// query
//========================================================================

struct TheAlgorithm
{
#ifdef NEW_ALGORITHM
	struct RunResult : mem_man<RunResult>, map<mgState,InXMLSet*> {};
#else
	typedef map<mgState, InXMLSet*> StateMap;
	struct RunResult : mem_man<RunResult>, vector<StateMap>
		{ RunResult(int N) : vector<StateMap>(1<<N){} };
#endif

	mem_man<RunResult>::cleaner rmmsc;

	Automaton& a;
	Context&   ctx;
	int        N;
	RunResult& NullResult;

	TheAlgorithm(Automaton& a, Context& ctx)
		: a(a), ctx(ctx), N(a.realVarName.size()),
#ifdef NEW_ALGORITHM
		 NullResult(*new RunResult)
#else
		 NullResult(*new RunResult(N))
#endif
	{
#ifdef NEW_ALGORITHM
		NullResult[a.init()] = 0;
#else
		NullResult[0][a.init()] = 0;
#endif
	}

	RunResult& bottomUp( InXML* ix )
	{
		// recursive version
		RunResult& L = ix->L ? bottomUp(ix->L) : NullResult;
		RunResult& R = ix->R ? bottomUp(ix->R) : NullResult;
		return *bottomUp_step(&L, ix, &R);
	}

	RunResult& bottomUp_s( InXML* root )
	{
		// stack version
		stack< pair<InXML*,RunResult*> > stk;
		stk.push( make_pair(root, (RunResult*)0) );

		for(;;) {
			pair<InXML*,RunResult*> t = stk.top();
			if( t.first )
				if( t.second )
					stk.push( make_pair(t.first->R, (RunResult*)0) );
				else
					stk.push( make_pair(t.first->L, (RunResult*)0) );
			else {
				RunResult* r = &NullResult;
				stk.pop();
				for(;;)
				{
					if( stk.empty() )
						return *r;
					pair<InXML*,RunResult*>& t = stk.top();
					if( t.second )
					  { r = bottomUp_step(t.second, t.first, r); stk.pop(); }
					else
					  { t.second = r; break; }
				}
			}
		}
	}

	RunResult* bottomUp_step( RunResult* L, InXML* ix, RunResult* R )
	{
#ifdef NEW_ALGORITHM
		RunResult* Ans = new RunResult;

		vector<char> A(a.varNum()); ix->set_flags( a, &A );

		SET_T mask = 0;
		for(int i=0; i!=N; ++i)
			mask |= ctx.never_bind(a.realVarName[i], ix)<<i;

		for(RunResult::iterator il=L->begin(); il!=L->end(); ++il)
		for(RunResult::iterator ir=R->begin(); ir!=R->end(); ++ir)
		{
			mgState   qL = il->first,           qR = ir->first;
			InXMLSet *sL = il->second,         *sR = ir->second;
			SET_T     mL = a.mask_of_state[qL], mR = a.mask_of_state[qR];
			if( !(mL & mR) )
				for(mgState qC=0; qC<a.size(); ++qC)
					if( a.mask_of_state.count(qC) )
					{
						SET_T mC = a.mask_of_state[qC];
						if( (mL|mR) == ((mL|mR)&mC) )
						{
							SET_T m = mC & ~(mL|mR);
							if( (m&mask)==0 )
							{
								for(int i=0; i!=N; ++i)
									A[a.realVarMap[i]] = (m & (1<<i) ? 1 : 0);
								if( qC == a.delta( qL, qR, &A[0] ) )
									(*Ans)[qC] = Union((*Ans)[qC], Prod(mL,sL,mC,Single(m,ix),mR,sR));
							}
						}
					}
		}

		return Ans;
#else
		RunResult* Ans = new RunResult(N);

		vector<char> A(a.varNum()); ix->set_flags( a, &A );

		SET_T mask = 0;
		for(int i=0; i!=N; ++i)
			mask |= ctx.never_bind(a.realVarName[i], ix)<<i;

		for(SET_T qi=0; qi<(1<<N); ++qi)
		for(SET_T ci=0; ci<=qi   ; ++ci) if( (ci&qi)==ci ) if((mask&ci) == 0)//ctx_mask
		{
			for(int i=0; i!=N; ++i)
				A[a.realVarMap[i]] = (ci & (1<<i) ? 1 : 0);

			for(SET_T li=0; li<=qi; ++li) if( (li&~ci&qi) == li ) {
				SET_T ri = qi&~ci&~li;
				StateMap &Lmap=(*L)[li], &Rmap=(*R)[ri];
				for(StateMap::iterator il=Lmap.begin(); il!=Lmap.end(); ++il)
				for(StateMap::iterator ir=Rmap.begin(); ir!=Rmap.end(); ++ir) {
					mgState q = a.delta( il->first, ir->first, &A[0] );
					(*Ans)[qi][q] = Union((*Ans)[qi][q],
						Prod(li,il->second,ci,Single(ci,ix),ri,ir->second));
				}
			}
		}

		return Ans;
#endif
	}

	void run( vector< vector<InXML*> >& result )
	{
		InXMLSet* ans = 0;
#ifdef NEW_ALGORITHM
		RunResult& sm = bottomUp_s(ctx.root);
		for(RunResult::iterator i=sm.begin(); i!=sm.end(); ++i)
#else
		StateMap& sm = bottomUp_s(ctx.root)[(1<<N)-1];
		for(StateMap::iterator i=sm.begin(); i!=sm.end(); ++i)
#endif
			if( a.is_final(i->first) )
				ans = Union(ans, i->second);
		if( ans )
			ans->as_vec( N, result );
		sort( result.begin(), result.end(), InXMLCmp() );

		vector< vector<InXML*> > r2;
		for(int i=0; i!=result.size(); ++i) {
			for(int j=0; j!=a.pairedVars.size(); ++j)
				if( result[i][a.pairedVars[j].first]->D
				 != result[i][a.pairedVars[j].second]->D )
					goto bad;
			r2.push_back(result[i]);
			bad:;
		}
		r2.swap(result);
	}
};

vector< vector<InXML*> >& nary_query( const string& v, Automaton& a, Context& ctx, QueryCache& qc )
{
	if( !qc.query_result.count(&a) ) {
		vector< vector<InXML*> >& result = qc.query_result[&a];
		if( !a.is_tautology() )
			TheAlgorithm(a,ctx).run(result);

		if( a.varMap.count(v) ) {
			set<InXML*>& pb = qc.possibly_bound[&a];
			int i = a.varMap[v];
			for(int j=0; j!=result.size(); ++j)
				pb.insert( result[j][i] );
		} else {
			// ToDo: true/falsȅꍇ
			// ToDo: possibly_bound̍XV
			// falsȅꍇpossibly_boundǂȂ̂okBtruȅꍇH
		}
	}
	return qc.query_result[&a];
}

void Engine::set_all_nodes( vector<InXML*>& qr, InXML* root )
{
	qr.push_back(root);
	for(InXML* p=root->L; p; p=p->R)
		set_all_nodes(qr, p);
}

bool Engine::query_list( vector<InXML*>& qr, const string& v,
                 Automaton& a, Context& ctx, QueryCache& qc )
{
	vector< vector<InXML*> >& result = nary_query( v, a, ctx, qc );
	if( a.is_tautology() )
		return false;

	// ݂bind(ctx)gresult񕪒Tōi
	// ɂȂ              ==> Empty
	// ͈͂͂邪
	//   - a.varMap includes V ==> ͈̔
	//   - not include         ==> All [false]Ԃ
	vector<InXML*> bound(a.realVarName.size());
	for(int i=0; i!=bound.size(); ++i)
		if( v!=a.realVarName[i] && a.realVarName[i].find("__q")!=0 )
			bound[i] = ctx[a.realVarName[i]];
	pair<vector< vector<InXML*> >::iterator,vector< vector<InXML*> >::iterator >
		rng = equal_range( result.begin(), result.end(), bound, InXMLCmp() );
	if( rng.first==rng.second )
		return true;
	if( a.varMap.count(v) ) {
		int j = a.invRealVarMap[a.varMap[v]];
		for(vector< vector<InXML*> >::iterator i=rng.first; i!=rng.second; ++i)
			qr.push_back( (*i)[j] );
		// ToDo uniqueȂƂ񂭂Ȃ
		return true;
	}
	return false; // All
}

//========================================================================
// interpreter
//========================================================================

vector<OutXML*> Engine::execute( InXML* root )
{
	mem_man<InXMLSet>::cleaner smmsc;
	Context ctx(root);
	vector<OutXML*> result;
	eval( result, main_template, ctx );
	return result;
}
