#include "common.h"

//========================================================================
// InXML Impls
//========================================================================

struct InXML_Labeled : InXML
{
	InXML_Labeled( int id, const string& d, const string& t )
		: InXML(id,d), tag(t+d), tagG(t) {
		for(size_t i=0;;i+=6) {
			i = tag.find(':', i);
			if( i == string::npos )
				break;
			tag.replace(i, 1, "_cOLOn_");
		}
	}
	string tag, tagG;

	void set_flags( const Automaton& a, vector<char>* A ) {
		map<string,int>::const_iterator p  = a.varMap.find(tag);
		map<string,int>::const_iterator pg = a.varMap.find(tagG);
		if(p !=a.varMap.end()) (*A)[p ->second] = 1;
		if(pg!=a.varMap.end()) (*A)[pg->second] = 1;
	}
};

struct InXML_Element : InXML_Labeled
{
	InXML_Element( int id, const string& d ) : InXML_Labeled(id,d,"__e") {}

	OutXML* convToOx() {
		vector<OutXML*> C;
		for(InXML* p=L; p; p=p->R)
			C.push_back( p->as_OutXML() );
		OutXML* ox = new OutXML_Element( D, C );
		ox->ix = this;
		return ox;
	}
};

struct InXML_Attribute : InXML_Labeled
{
	InXML_Attribute( int id, const string& d ) : InXML_Labeled(id,d,"__a") {}

	OutXML* convToOx() {
		vector<OutXML*> C;
		for(InXML* p=L; p; p=p->R)
			C.push_back( p->as_OutXML() );
		OutXML* ox = new OutXML_Attribute( D, C );
		ox->ix = this;
		return ox;
	}
};

struct InXML_Text : InXML
{
	InXML_Text( int id, const string& d ) : InXML(id,d) {}

	OutXML* convToOx() {
		OutXML* ox = new OutXML_Text( D );
		ox->ix = this;
		return ox;
	}

	void set_flags( const Automaton& a, vector<char>* A ) {
		map<string,int>::const_iterator p  = a.varMap.find("__t");
		if(p !=a.varMap.end()) (*A)[p ->second] = 1;
	}
};

//========================================================================
// InXML Loader
//========================================================================

struct InXML_Loader : SAXHandler
{
	InXML_Loader() : id(0) {}

	int id;
	vector<InXML*> stk;

	bool start_e( const char* qname, const vector<SAXAttr>& sas )
	{
		InXML* e = new InXML_Element(id++, qname);

		if( !stk.empty() )
			stk.back()->append(e);
		stk.push_back(e);

		for(int i=0; i!=sas.size(); ++i) {
			InXML* a = new InXML_Attribute( id++, sas[i].qname );
			InXML* t = new InXML_Text(id++, sas[i].value);
			a->append(t);
			e->append(a);
		}
		return true;
	}

	bool end_e()
	{
		if(stk.size() != 1)
			stk.pop_back();
		return true;
	}

	bool text( const vector<char>& data )
	{
		InXML* t = new InXML_Text(id++, string(data.begin(), data.end()));
		stk.back()->append(t);
		return true;
	}
};


InXML* loadXML( const char* xmlfile )
{
	InXML_Loader ixl;
	return parse( xmlfile, &ixl ) ? ixl.stk[0] : 0;
}
