
struct  InXML;
struct OutXML;

//========================================================================
// Output XML Nodes
//========================================================================

struct OutXML : mem_man<OutXML>
{
// Tree Members
	string          D;
	vector<OutXML*> C;

// InXML form
	InXML* ix;
	InXML* as_InXML() { return ix; }

// methods
	virtual void output( ostream& os ) = 0;
	virtual void output_text( ostream& os ) {}
	virtual ~OutXML() {}
	virtual OutXML* clone_with( vector<OutXML*>& c ) = 0;

// construction
	OutXML() : ix(0) {}
};

inline ostream& operator<<( ostream& os, OutXML& ox ) {
	ox.output(os); return os;
}

//========================================================================
// OutXML Impls
//========================================================================

struct OutXML_Text : OutXML
{
	OutXML_Text( const string& d ) {D=d;}

	void output( ostream& os ) {
		for(int i=0; i!=D.size(); ++i)
			switch( D[i] ) {
			case '&': os << "&amp;"; break;
			case '<': os << "&lt;";  break;
			case '>': os << "&gt;";  break;
			default:  os << D[i];    break;
			}
	}

	void output_text( ostream& os ) {
		for(int i=0; i!=D.size(); ++i)
			switch( D[i] ) {
			case '&': os << "&amp;"; break;
			case '<': os << "&lt;";  break;
			case '>': os << "&gt;";  break;
			case '\n':os << " ";     break;
			default:  os << D[i];    break;
			}
	}

	OutXML* clone_with( vector<OutXML*>& c ) { return this; }
};

struct OutXML_Attribute : OutXML
{
	OutXML_Attribute( const string& d, const vector<OutXML*> c ) {D=d; C=c;}

	void output( ostream& os ) {
		os << D << "=\"";
		for(int i=0; i!=C.size(); ++i) {
			C[i]->output_text(os);
		}
		os << "\"";
	}

	OutXML* clone_with( vector<OutXML*>& c ) {
		OutXML* n = new OutXML_Attribute(D, c); n->ix=ix; return n; }
};

struct OutXML_Element : OutXML
{
	OutXML_Element( const string& d, const vector<OutXML*> c ) {D=d; C=c;}

	void output( ostream& os ) {
		os << "<" << D;
		bool empty = true;
		for(int i=0; i!=C.size(); ++i) {
			if( dynamic_cast<OutXML_Attribute*>(C[i]) )
				os << " " << *C[i];
			else
				empty = false;
		}
		if( empty )
		{
			os << "/>";
			return;
		}
		os << ">";
		for(int i=0; i!=C.size(); ++i)
			if( dynamic_cast<OutXML_Attribute*>(C[i]) ) {} else
				os << *C[i];
		os << "</" << D << ">";
	}

	void output_text( ostream& os ) {
		for(int i=0; i!=C.size(); ++i)
			C[i]->output_text(os);
	}

	OutXML* clone_with( vector<OutXML*>& c ) {
		OutXML* n = new OutXML_Element(D, c); n->ix=ix; return n; }
};

//========================================================================
// Input XML Nodes
//========================================================================

struct InXML : mem_man<InXML>
{
// Tree Members
	string  D;
	InXML*  L;
	InXML*  R;

// Constructions
	int           Id;
	InXML* LastChild;
	InXML(int id, const string& d) : Id(id), D(d), L(0), R(0), ox(0) {}
	void append( InXML* p ) { LastChild = (L ? (LastChild->R=p) : (L=p)); }

// OutXML form
	OutXML* ox;
	OutXML* as_OutXML() { return ox ? ox : (ox=convToOx()); }

// methods
	virtual OutXML* convToOx() = 0;
	virtual void set_flags( const Automaton& a, vector<char>* A ) {}
	virtual ~InXML() {}

	bool is_anscestor_or_self_of( InXML* rhs ) {
		InXML* LC = this;
		while( LC->L ) LC=LC->LastChild;
		return Id <= rhs->Id && rhs->Id <= LC->Id;
	}
};

struct InXMLCmp
{
	bool operator()(const vector<InXML*>& lhs, const vector<InXML*>& rhs) const
	  { return lexicographical_compare(lhs.begin(), lhs.end(),
	                                   rhs.begin(), rhs.end(), *this); }
	bool operator()( InXML* lhs, InXML* rhs ) const
	  {return !lhs?false:!rhs?false:lhs->Id<rhs->Id;}
};

InXML* loadXML( const char* xmlfile );
